import asyncio
import random
import traceback
from collections import namedtuple
from importlib import import_module

import rx
import rx.operators as ops
from rx.scheduler.eventloop import AsyncIOScheduler

import rxx

from cyclotron.backpressure import pid
from cyclotron.debug import trace_observable
import cyclotron_aiokafka as kafka

from .driver import app_sink, app_source
from .util import import_function


def initialize_topics(config_topics):
    ''' initializes the topic objects from a configuration

    If no encoding os provided, then the string encoder is used by default.
    '''
    Topic = namedtuple('Topic', [
        'name',
        'encode', 'decode',
        'map_partition',
        'start_from', 'timestamp_mapper',
        'merge_lookup_depth',
    ])

    pull_mode = False
    topics = {}
    for topic in config_topics:
        if "encoder" in topic:
            module = import_module(topic['encoder'])
        else:
            module = import_module("makinage.encoding.string")
        encoder = getattr(module, "encoder")
        encode, decode = encoder()

        if "partition_selector" in topic:
            map_partition = import_function(topic['partition_selector'])
        else:
            def r(i): return int(random.random() * 1000)
            map_partition = r

        if "start_from" in topic:
            start_from = topic['start_from']
        else:
            start_from = 'end'

        if "timestamp_mapper" in topic:
            timestamp_mapper = import_function(topic['timestamp_mapper'])
            pull_mode = True
        else:
            timestamp_mapper = None

        merge_lookup_depth = 1
        if "merge_lookup_depth" in topic:
            merge_lookup_depth = topic['merge_lookup_depth']

        topics[topic['name']] = Topic(
            name=topic['name'],
            encode=encode,
            decode=decode,
            map_partition=map_partition,
            start_from=start_from,
            timestamp_mapper=timestamp_mapper,
            merge_lookup_depth=merge_lookup_depth,
        )

    return topics, pull_mode


def subscribe_to_non_regulated_sinks(config, sinks_feedback):
    """Ensures that all feedback observable are subscribed

    The app sinks start consuming when their feedback observable is subscribed.
    So we subscribe to all sinks that are not regulated.
    """
    if 'sinks' not in config:
        return

    regulated_sinks = []
    if 'regulators' in config:
        for regulator in config['regulators']:
            regulated_sinks.append(regulator['feedback'])

    sinks_feedback.pipe(
        ops.filter(lambda i: i.id not in regulated_sinks),
        ops.flat_map(lambda i: i.observable),
    ).subscribe()


def initialize_regulators(config, kafka_feedback, app_feedback):
    regulators = {}
    app_regulators = []
    for regulator in config['regulators']:
        app = False
        if 'sinks' in config and regulator['feedback'] in config['sinks']:  # app sink feedback
            app = True
            feedback = app_feedback.pipe(
                ops.filter(lambda i: i.id == regulator['feedback']),
                ops.flat_map(lambda i: i.observable),
            )
        else:  # kafka sink feedback
            feedback = kafka_feedback.pipe(
                ops.filter(lambda i: i[0] == regulator['feedback']),
            )

        control = feedback.pipe(
            ops.map(lambda i: i[1] / 1000),
            pid(rx.concat(rx.just(1.0), rx.never()), -0.001, -0.001, 0.0),
            ops.map(lambda i: max(min(i, 0.01), 0.0)),
            ops.filter(lambda i: i > 0),
            ops.share(),
        )

        regulators[regulator['control']] = control
        if app:
            app_regulators.append(control)

    for app_control in app_regulators:
        app_control.subscribe()

    print(regulators)
    return regulators


def create_kafka_source_observable(kafka_source, topic):
    if topic.timestamp_mapper is not None:  # pull mode
        return kafka_source.pipe(
            ops.filter(lambda i: i.topic == topic.name),  # ConsumerRecords
            ops.flat_map(lambda i: i.records.pipe(
                rxx.pullable.sorted_merge(
                    key_mapper=topic.timestamp_mapper,
                    lookup_size=topic.merge_lookup_depth,
                ),
                rxx.pullable.push(),
            )),
        )
    else:  # push mode
        return kafka_source.pipe(
            trace_observable("kafka tst"),
            ops.filter(lambda i: i.topic == topic.name),  # ConsumerRecords
            ops.flat_map(lambda i: i.records.pipe(
                ops.merge_all(),
            )),
        )


def create_app_source_observable(source, source_id):
    return source.pipe(
        ops.filter(lambda i: i.id == source_id),  # ConsumerRecords
        ops.flat_map(lambda i: i.observable),
    )


def create_operators(config, config_source,
                     kafka_source, kafka_feedback,
                     app_source_data, app_feedback):
    ''' creates the operators declared in config

    Args:
        config: a dict containing the initial configuration file.
        config_source: An observable emitting configuration items.
        kafka_source: The kafka response observable
        kafka_feedback: The kafka backpressure process feedback
    Returns:
        An observable containing tuples of (topic, observable).
    '''
    try:
        source_type = kafka.DataSourceType.STREAM
        if "source_type" in config['application']:
            source_type = kafka.DataSourceType.BATCH if config['application']['source_type'] == "batch" else kafka.DataSourceType.STREAM

        topics, pull_mode = initialize_topics(config['topics'])
        datafeed_mode = kafka.DataFeedMode.PULL if pull_mode is True else kafka.DataFeedMode.PUSH
        if 'regulators' in config:
            regulators = initialize_regulators(
                config,
                kafka_feedback, app_feedback,
            )
        else:
            regulators = {}

        subscribe_to_non_regulated_sinks(config, app_feedback)

        producers = []
        consumers = []
        source_connectors = []
        sink_connectors = []
        for k, operator in config['operators'].items():
            factory = import_function(operator['factory'])
            sources = []
            if 'sources' in operator:
                for source in operator['sources']:
                    print('create source {}'.format(source))
                    if 'sources' in config and source in config['sources']:  # This is an app connector
                        source_config = config['sources'][source]
                        source_kwargs = source_config.get('kwargs', None)
                        source_factory = import_function(source_config['factory'], source_kwargs)
                        source_observable = source_factory()
                        source_connectors.append(app_source.Create(
                            id=source,
                            observable=source_observable,
                            control=regulators[source] if source in regulators else None,
                        ))

                        sources.append(create_app_source_observable(
                            app_source_data, source
                        ))
                    else:  # A kafka source
                        consumers.append(kafka.ConsumerTopic(
                            topic=source,
                            decode=topics[source].decode,
                            control=regulators[source] if source in regulators else None,
                            start_from=topics[source].start_from,
                        ))

                        sources.append(create_kafka_source_observable(
                            kafka_source, topics[source],
                        ))

            print(sources)
            sinks = factory(config_source, *sources)
            print("sinks: {}".format(sinks))
            if 'sinks' in operator:
                for index, sink in enumerate(operator['sinks']):
                    print('create sink {} at {}'.format(sink, index))
                    if 'sinks' in config and sink in config['sinks']:  # This is an app connector
                        sink_config = config['sinks'][sink]
                        sink_kwargs = sink_config.get('kwargs', None)
                        sink_factory = import_function(sink_config['factory'], sink_kwargs)
                        sink_operator = sink_factory()
                        sink_connectors.append(app_sink.Create(
                            id=sink,
                            operator=sink_operator,
                            observable=sinks[index],
                        ))
                    else:  # Kafka
                        producers.append(kafka.ProducerTopic(
                            topic=sink,
                            records=sinks[index],
                            map_key=lambda i: None,
                            encode=topics[sink].encode,
                            map_partition=topics[sink].map_partition
                        ))

        kafka_sink = []
        if len(consumers) > 0:
            kafka_sink.append(kafka.Consumer(
                server=config['kafka']['endpoint'],
                group=config['application']['name'],
                topics=rx.from_(set(consumers)),
                source_type=source_type,
                feed_mode=datafeed_mode,
            ))

        if len(producers) > 0:
            kafka_sink.append(kafka.Producer(
                server=config['kafka']['endpoint'],
                topics=rx.from_(producers)
            ))

        kafka_sink = rx.from_(kafka_sink) if len(kafka_sink) > 0 else rx.never()
        source_connectors = rx.from_(source_connectors) if len(source_connectors) > 0 else rx.never()
        sink_connectors = rx.from_(sink_connectors) if len(sink_connectors) > 0 else rx.never()

        return rx.just((kafka_sink, source_connectors, sink_connectors))

    except Exception as e:
        print("Error while creating operators: {}, {}".format(
              e, traceback.format_exc()))



import asyncio
import pytest
import time

import rx
import rx.operators as ops
from rx.subject import Subject
from rx.scheduler.eventloop import AsyncIOScheduler
import rxsci as rs

from makinage.driver.app_source import make_driver, Sink, Create, create_controllable_source


@pytest.mark.asyncio
async def test_controllable_source():
    actual_result = []
    actual_error = []
    actual_completed = []
    actual_sleep = []

    def source():
        for i in range(3):
            yield i
            time.sleep(0.3)

    def test_sleep(i): actual_sleep.append(i)

    source = rx.from_(source())  # Subject()
    control = Subject()


    #async def test():
    controllable = create_controllable_source(
        source, control,
        loop=asyncio.get_event_loop(), sleep=test_sleep,
    )
    controllable.subscribe(
        on_next=actual_result.append,
        on_error=actual_error.append,
        on_completed=lambda: actual_completed.append(True),
    )

    await asyncio.sleep(0.2)
    assert actual_result == [0]
    control.on_next(1.0)
    await asyncio.sleep(0.2)
    assert actual_sleep == [1.0]
    await asyncio.sleep(0.4)
    assert actual_result == [0, 1, 2]

    control.on_completed()
    await asyncio.sleep(0.5)
    assert actual_completed == [True]


@pytest.mark.asyncio
async def test_driver():
    actual_result = []
    actual_error = []
    actual_completed = []
    actual_sleep = []

    def test_sleep(i): actual_sleep.append(i)

    def source():
        for i in range(3):
            yield i
            time.sleep(0.3)

    source = rx.from_(source())
    control = Subject()
    component = make_driver(loop=asyncio.get_event_loop(), sleep=test_sleep)

    controllable_source = component.call(Sink(
        connector=rx.just(Create('test', source, control))
    ))

    controllable_source.source.pipe(
        rs.ops.assert_(lambda i: i.id == 'test'),
        ops.flat_map(lambda i: i.observable),
    ).subscribe_(
        on_next=actual_result.append,
        on_error=actual_error.append,
        on_completed=lambda: actual_completed.append(True),
    )

    await asyncio.sleep(0.2)
    assert actual_result == [0]
    control.on_next(1.0)
    await asyncio.sleep(0.2)
    assert actual_sleep == [1.0]
    print("fizz")
    await asyncio.sleep(0.5)
    assert actual_result == [0, 1, 2]

    control.on_completed()
    await asyncio.sleep(0.2)
    assert actual_completed == [True]



from makinage.util import import_function
from makinage.sample.functions import fn_no_param


def test_import():
    fn = import_function("makinage.sample.functions:fn_no_param")

    assert fn is fn_no_param
    assert fn() == 1


def test_import_kwargs():
    fn = import_function("makinage.sample.functions:fn_add", kwargs={"first": 1, "second": 2})

    assert fn() == 3


def test_import_kwargs_override():
    fn = import_function("makinage.sample.functions:fn_add_with_defaults", kwargs={"first": 3})

    assert fn() == 5
    assert fn(second=5) == 8
    assert fn(first=2) == 4



from typing import Callable
from collections import namedtuple
import queue
import asyncio

import rx
import rx.operators as ops
from rx.core.notification import OnNext, OnError
from rx.scheduler import NewThreadScheduler
from rx.scheduler.eventloop import AsyncIOThreadSafeScheduler

from cyclotron import Component


def QObservable(q):
    def on_subscribe(observer, scheduler):
        while(True):
            i = q.get()
            if isinstance(i, OnNext):
                observer.on_next(i.value)
                q.task_done()
            elif isinstance(i, OnError):
                observer.on_error(i.exception)
                break
            else:
                observer.on_completed()
                break

    return rx.create(on_subscribe)


def create_controllable_sink(sink_id: str,
                             operator: Callable[[rx.Observable], rx.Observable],
                             sink: rx.Observable,
                             ) -> rx.Observable:
    """Creates a controllable sink from a sink operator

    The sink operator must subscribe to its provided observable. All items are
    emitted on a thread dedicated to this operator. So it can do
    blocking calls without issues on the makinage eventloop.

    Args:
        - sink_id: The observable identifier used in the feedback loop.
        - operator: The connector logic to isolate.
        - sink: A function that subscribes to an observable.
    """
    q = queue.Queue()
    index = 0

    lifecycle = QObservable(q).pipe(
        ops.subscribe_on(NewThreadScheduler()),
        operator,
        ops.ignore_elements(),
        ops.observe_on(AsyncIOThreadSafeScheduler(asyncio.get_event_loop())),
    )

    def push_item(q, sink_id):
        def _push_item(i):
            nonlocal index
            q.put_nowait(i)
            if isinstance(i, OnNext):
                index += 1
                if index == 500:
                    index = 0
                    return (sink_id, q.qsize())
            return None
        return ops.map(_push_item)

    feedback = sink.pipe(
        ops.materialize(),
        push_item(q, sink_id),
        ops.filter(lambda i: i is not None)
    )

    return rx.merge(lifecycle, feedback)


Sink = namedtuple('Sink', ['connector'])
Source = namedtuple('Source', ['feedback'])

# Sink items
Create = namedtuple('Create', ['id', 'operator', 'observable'])
Create.__doc__ = "Creates a new application defined sink."
Create.id.__doc__ = "The id of the sink."
Create.operator.__doc__ = "The function implemententing the sink logic, as an operator."
Create.observable.__doc__ = "The observable that will feed the sink."


# Source items
Feedback = namedtuple('Feedback', ['id', 'observable'])
Feedback.__doc__ = "The feedback loop of a sink."
Feedback.id.__doc__ = "The id of the sink."
Feedback.observable.__doc__ = "The feedback observable, emitting (id, qsize) items."


def make_driver():

    def driver(sink):
        """ Application sink driver.

        This is a driver implementation where the logic is pushed by
        the application. This allows to write connectors in a more flexible way
        than writing a driver per connector.

        The Sink connector stream emits the following items: Create
        The Source feedback stream emits the following items: Feedback
        """
        def on_subscribe(observer, scheduler):
            def on_next(i):
                if type(i) is Create:
                    observer.on_next(
                        Feedback(
                            i.id,
                            create_controllable_sink(
                                i.id,
                                i.operator,
                                i.observable,
                            )
                        )
                    )

                else:
                    observer.on_error("app sink unknown command: {}".format(i))

            disposable = sink.connector.subscribe(
                on_next=on_next,
                on_error=observer.on_error,
                on_completed=observer.on_completed)

            return disposable

        return Source(
            feedback=rx.create(on_subscribe),
        )

    return Component(call=driver, input=Sink)



import os
import pytest
from collections import namedtuple

from rx.subject import Subject

pytest.importorskip("makinage.serve.serve")
from makinage.serve.serve import create_transform_functions, \
    create_model_predict, infer, serve
from makinage.sample.serve import ZeroModel, predict_zero

import numpy as np

zero_model_dirname = os.path.join('assets', 'zero_mlflow_pyfunc.zip')


def test_create_transform_default():
    c = {'config': {'serve': {}}}
    t = create_transform_functions(c)
    assert t.pre == np.array
    assert t.post(1, 1) == (1, 1)


def test_create_predict_default():
    c = {'config': {'serve': {}}}
    model = ZeroModel()
    p = create_model_predict(model, c)
    assert p is predict_zero


def test_create_predict_custom():
    c = {'config': {'serve': {
        'predict': 'makinage.sample.serve:predict',
        'ratio': 2,
    }}}

    model = ZeroModel()
    p = create_model_predict(model, c)
    assert p(2) == (0.0, 4)


def test_predict():
    predict_count = 0

    def _predict(i):
        nonlocal predict_count
        predict_count += 1
        return np.array([i+1])

    c = {'config': {'serve': {
        'pre_transform': 'makinage.sample.serve:pre_transform',
        'post_transform': 'makinage.sample.serve:post_transform',
        'input_field': 'x', 'output_field': 'pred',
    }}}
    t = create_transform_functions(c)
    data = {'x': 42}
    actual_result = infer(data, t, _predict)

    assert predict_count == 1
    assert actual_result == {'x': 42, 'pred': [43]}


def test_serve():
    config = Subject()
    model = Subject()
    data = Subject()
    prediction, = serve(config, model, data)

    actual_predictions = []
    prediction.subscribe(on_next=actual_predictions.append)

    config.on_next({'config': {'serve': {
        'predict': 'makinage.sample.serve:predict',
        'ratio': 2,
    }}})

    with open(zero_model_dirname, 'rb') as fd:
        model_archive_data = fd.read()
    model.on_next(model_archive_data)

    data.on_next(1)
    assert actual_predictions == [
        (1, (0.0, 2))
    ]

    # update config
    actual_predictions.clear()
    config.on_next({'config': {'serve': {
        'predict': 'makinage.sample.serve:predict',
        'ratio': 3,
    }}})

    data.on_next(1)
    assert actual_predictions == [
        (1, (0.0, 3))
    ]



from collections import namedtuple
import time
import asyncio

import rx
import rx.operators as ops
from rx.scheduler import NewThreadScheduler
from rx.scheduler.eventloop import AsyncIOThreadSafeScheduler

from cyclotron import Component


ControlItem = namedtuple('ControlItem', ['delay'])


def control_sync(i, sleep):
    if type(i) is ControlItem:
        sleep(i.delay)
        return ControlItem
    return i


def create_controllable_source(source, control, loop, sleep):
    """Makes an observable controllable to handle backpressure

    This function takes an observable as input makes it controllable by
    executing it in a dedicated worker thread. This allows to regulate
    the emission of the items independently of the asyncio event loop.

    Args:
        - source: An observable emitting the source items.
        - control: [Optional] The control observable emitting delay items in seconds.
        - sleep: the sleep function to used. Needed only for testing.

    Returns:
        An observable similar to the source observable, with emission being
        controlled by the control observable.
    """
    if control is not None:
        typed_control = control.pipe(
            ops.observe_on(NewThreadScheduler()),
            ops.map(ControlItem),
        )

        scheduled_source = source.pipe(
            ops.subscribe_on(NewThreadScheduler()),
            ops.merge(typed_control),
            ops.map(lambda i: control_sync(i, sleep)),
            ops.filter(lambda i: i is not ControlItem),
            ops.observe_on(AsyncIOThreadSafeScheduler(loop)),
        )
    else:
        scheduled_source = source.pipe(
            ops.subscribe_on(NewThreadScheduler()),
            ops.observe_on(AsyncIOThreadSafeScheduler(loop)),
        )

    return scheduled_source


Sink = namedtuple('Sink', ['connector'])
Source = namedtuple('Source', ['source'])

# Sink items
Create = namedtuple('Create', ['id', 'observable', 'control'])
Create.__doc__ = "Creates a new application defined source."
Create.id.__doc__ = "The id of the source."
Create.observable.__doc__ = "The observable that will feed the source."
Create.control.__doc__ = "The control observable used to regulate data emission."


# Source items
ControllableSource = namedtuple('ControllableSource', ['id', 'observable'])
ControllableSource.__doc__ = "The regulated observable of a source."
ControllableSource.id.__doc__ = "The id of the source."
ControllableSource.observable.__doc__ = "The observable emitting source items."


def make_driver(sleep=time.sleep, loop=None):
    loop = loop or asyncio.get_event_loop()

    def driver(sink):
        def on_subscribe(observer, scheduler):
            _scheduler = scheduler or AsyncIOThreadSafeScheduler(loop)

            def on_next(i):
                if type(i) is Create:
                    observer.on_next(
                        ControllableSource(
                            i.id,
                            create_controllable_source(
                                source=i.observable,
                                control=i.control,
                                loop=loop,
                                sleep=sleep,
                            )
                        )
                    )

                else:
                    observer.on_error("app source unknown command: {}".format(i))

            disposable = sink.connector.subscribe(
                on_next=on_next,
                on_error=lambda e: _scheduler.schedule(lambda s, t: observer.on_error(e)),
                on_completed=lambda: _scheduler.schedule(lambda s, t: observer.on_completed()),
            )

            return disposable

        return Source(source=rx.create(on_subscribe))

    return Component(call=driver, input=Sink)



import asyncio
import pytest
import queue
import traceback

import rx
import rx.operators as ops
from rx.subject import Subject
import rxsci as rs
from makinage.driver.app_sink import make_driver, Sink, Create, create_controllable_sink


def on_error(e):
    raise e

@pytest.mark.asyncio
async def test_controllable_sink():
    actual_result = []
    feedback_result = []
    feedback_error = []
    feedback_completed = []

    sink = Subject()
    sink_lock_q = queue.Queue()

    def op(source):
        return source.pipe(
            ops.do_action(lambda i: actual_result.append(i)),
            # block the connector's operator thread
            ops.do_action(lambda i: sink_lock_q.get()),
        )

    feedback = create_controllable_sink('test', op, sink)

    feedback.subscribe(
        on_next=feedback_result.append,
        on_error=feedback_error.append,
        on_completed=feedback_completed.append,
    )

    for i in range(450):
        sink.on_next(i)

    await asyncio.sleep(0.1)
    assert feedback_result == []
    assert actual_result == [0]

    for i in range(10):
        sink_lock_q.put_nowait(i)

    await asyncio.sleep(0.1)
    assert feedback_result == []
    assert actual_result == list(range(11))

    for i in range(50):
        sink.on_next(i)

    assert feedback_result == [('test', 489)]


@pytest.mark.asyncio
async def test_driver():
    sink_lock_q = queue.Queue()
    sink_data = Subject()
    component = make_driver()

    actual_result = []
    feedback_result = []
    feedback_completed = []

    def test_operator():
        return rx.pipe(
            ops.do_action(print),
            ops.do_action(lambda i: actual_result.append(i)),
            # block the connector's operator thread
            ops.do_action(lambda i: sink_lock_q.get()),
        )

    source = component.call(Sink(
        connector=rx.just(Create('test', test_operator(), sink_data))
    ))

    source.feedback.pipe(
        ops.do_action(print),
        rs.ops.assert_(lambda i: i.id == 'test'),
        ops.flat_map(lambda i: i.observable),
    ).subscribe(
        on_next=feedback_result.append,
        on_error=on_error,
        on_completed=feedback_completed.append,
    )

    for i in range(450):
        sink_data.on_next(i)

    await asyncio.sleep(0.1)
    assert feedback_result == []
    assert actual_result == [0]

    for i in range(10):
        sink_lock_q.put_nowait(i)

    await asyncio.sleep(0.1)
    assert feedback_result == []
    assert actual_result == list(range(11))

    for i in range(50):
        sink_data.on_next(i)

    assert feedback_result == [('test', 489)]



import pickle
import mlflow.pyfunc


class ZeroWrapper(mlflow.pyfunc.PythonModel):
    def load_context(self, context):
        with open(context.artifacts["zero_model"], 'rb') as fd:
            self.model = pickle.load(fd)

    def predict(self, context, model_input):
        return self.model.predict(model_input)



from makinage.encoding.json import encoder

def test_json_encoder():
    encode, decode = encoder()

    assert encode({"foo": "bar", "biz": 42}) == b'{"foo": "bar", "biz": 42}'
    assert decode(b'{"foo": "bar", "biz": 42}') == {"foo": "bar", "biz": 42}





def encoder(encoding="utf-8"):
    ''' String encoder

    encodes strings using the encode/decode python functions

    Args:
        encoding: The string encoding to use.
    '''
    def _encode(i):
        return i.encode(encoding)

    def _decode(i):
        return i.decode(encoding)

    return _encode, _decode



import asyncio
from collections import namedtuple
from functools import partial

import rx
import rx.operators as ops
from rx.disposable import Disposable
from rx.scheduler.eventloop import AsyncIOScheduler

from cyclotron import Component
from cyclotron.debug import trace_observable
from cyclotron.asyncio.runner import run
import cyclotron_std.sys.argv as argv
import cyclotron_std.io.file as file
import cyclotron_aiohttp.http as http

from .driver import app_sink, app_source

from .config import read_config_from_args
from .operator import create_operators

import cyclotron_aiokafka as kafka

MakiNageSink = namedtuple('MakiNageSink', [
    'kafka', 'app_sink', 'app_source', 'http', 'file',
])
MakiNageSource = namedtuple('MakiNageSource', [
    'kafka', 'app_sink', 'app_source', 'http', 'file', 'argv',
])
MakiNageDrivers = namedtuple('MakiNageDrivers', [
    'kafka', 'app_sink', 'app_source', 'http', 'file', 'argv'
])

Values = namedtuple('Values', ['id', 'observable'])


def ref_count(subscribe_count=1, unsubscribe_count=0):
    connectable_subscription = None
    count = 0

    def _ref_count(source):
        def subscribe(observer, scheduler=None):
            nonlocal connectable_subscription
            nonlocal count
            count += 1
            should_connect = count == subscribe_count
            subscription = source.subscribe(observer, scheduler=scheduler)
            if should_connect:
                connectable_subscription = source.connect(scheduler)

            def dispose():
                nonlocal connectable_subscription
                nonlocal count
                subscription.dispose()
                count -= 1
                if count == unsubscribe_count:
                    connectable_subscription.dispose()

            return Disposable(dispose)

        return rx.create(subscribe)

    return _ref_count


def makinage(aio_scheduler, sources):
    def on_error(e):
        raise e

    config, read_request, http_request = read_config_from_args(
        sources.argv.argv,
        sources.file.response,
        sources.http.response,
        scheduler=aio_scheduler
    )

    first_config = rx.concat(config.pipe(ops.take(1),), rx.never())

    kafka_source = sources.kafka.response.pipe(
        trace_observable("kafka source1"),
        ops.replay(),
        ops.ref_count(),
        trace_observable("kafka source2"),
    )
    kafka_source.subscribe(on_error=on_error)

    app_sources = sources.app_source.source.pipe(
        trace_observable("app source1"),
        ops.replay(),
        ops.ref_count(),
        trace_observable("app source2"),
    )
    app_sources.subscribe(on_error=on_error)

    app_sinks_feedback = sources.app_sink.feedback
    operators_sinks = first_config.pipe(
        ops.flat_map(lambda i: create_operators(
            i, config,
            kafka_source,
            sources.kafka.feedback.pipe(ops.share()),
            app_sources, app_sinks_feedback,
        )),
        ops.subscribe_on(aio_scheduler),
        ops.publish(),
        ref_count(subscribe_count=3),
        #trace_observable("makinage"),
    )

    kafka_request = operators_sinks.pipe(
        ops.flat_map(lambda i: i[0]),
        trace_observable("kafka_request"),
    )
    source_connectors = operators_sinks.pipe(
        ops.flat_map(lambda i: i[1]),
        trace_observable("source_connectors"),
    )
    sink_connectors = operators_sinks.pipe(
        ops.flat_map(lambda i: i[2]),
        trace_observable("sink_connectors"),
    )

    '''
    config.pipe(ops.subscribe_on(aio_scheduler)).subscribe(
        on_next=print,
        on_error=print,
    )
    '''

    return MakiNageSink(
        file=file.Sink(request=read_request),
        http=http.Sink(request=http_request),
        app_source=app_source.Sink(connector=source_connectors),
        app_sink=app_sink.Sink(connector=sink_connectors),
        kafka=kafka.Sink(request=kafka_request),
    )


def main():
    loop = asyncio.get_event_loop()
    # loop.set_debug(True)
    aio_scheduler = AsyncIOScheduler(loop=loop)
    run(
        Component(
            call=partial(makinage, aio_scheduler),
            input=MakiNageSource),
        MakiNageDrivers(
            kafka=kafka.make_driver(),
            app_sink=app_sink.make_driver(),
            app_source=app_source.make_driver(),
            http=http.make_driver(),
            file=file.make_driver(),
            argv=argv.make_driver(),
        ),
        loop=loop,
    )



import functools
from importlib import import_module


def import_function(spec, kwargs=None):
    """Loads a function definition from a string specs

    Args:
        - spec: Function speciciation in the form "module:function"
        - kwargs: [Optional] a dict containing the kwargs

    Returns:
        A function object partialy applied with kwargs
    """
    spec = spec.split(':')
    module = spec[0]
    fn = spec[1]
    module = import_module(module)
    fn = getattr(module, fn)

    if kwargs is not None:
        fn = functools.partial(fn, **kwargs)
    return fn



from collections import namedtuple


def pre_transform(config):
    field = config['config']['serve']['input_field']

    def _transform(i):
        return i[field]

    return _transform


def post_transform(config):
    field = config['config']['serve']['output_field']

    def _transform(i, pred):
        i[field] = pred
        return i

    return _transform


def predict_zero(i):
    return [0.0]


ZeroModel = namedtuple('ZeroModel', ['predict'])
ZeroModel.__new__.__defaults__ = (predict_zero,)


def predict(model, config):
    ratio = config['config']['serve']['ratio']

    def _predict(i):
        print(type(model))
        return (model.predict(i)[0], i*ratio)

    return _predict



import asyncio
import functools
from collections import namedtuple


from cyclotron import Component
from cyclotron.asyncio.runner import run
from cyclotron.debug import trace_observable

import cyclotron_std.sys.argv as argv
import cyclotron_std.io.file as file
import cyclotron_aiohttp.http as http
import cyclotron_aiokafka as kafka


import rx
import rx.operators as ops
from rx.subject import Subject
from rx.scheduler.eventloop import AsyncIOScheduler

from makinage.config import read_config_from_args
from makinage.encoding.none import encoder as none_encoder
import makinage

ModelPublisherSource = namedtuple('ModelPublisherSource', ['kafka', 'http', 'file', 'argv'])
ModelPublisherSink = namedtuple('ModelPublisherSink', ['kafka', 'http', 'file'])
ModelPublisherDrivers = namedtuple('ModelPublisherDrivers', ['kafka', 'http', 'file', 'argv'])


def read_file(path):
    '''Temporary side effect until a correct file adapter is writen
    '''
    with open(path, mode="rb") as f:
        return f.read()


def create_model_readers(config):
    readers = {}

    for model in config['models']:
        readers[model['topic']] = rx.just(model['path']).pipe(
            ops.map(read_file),
        )

    return readers


def create_model_topics(config):
    topics = []
    readers = create_model_readers(config)

    encode_none, _ = none_encoder()

    for model in config['models']:
        topics.append(kafka.ProducerTopic(
            topic=model['topic'],
            records=readers[model['topic']],
            map_key=lambda i: None,
            encode=encode_none,
            map_partition=lambda i: i,
        ))

    return kafka.Producer(
        server=config['kafka']['endpoint'],
        topics=rx.from_(topics),
        max_request_size=32*1024*1024,
    )


def model_publisher(scheduler, sources):
    file_source = sources.file.response.pipe(ops.share())

    # kafka driver bootstrap. fixme
    kafka_source = sources.kafka.response.pipe(
        ops.do_action(print),
        ops.replay(),
        ops.ref_count(),
    )
    kafka_source.subscribe()

    config, config_read_request, http_request = read_config_from_args(
        sources.argv.argv,
        file_source,
        sources.http.response,
        scheduler=scheduler
    )

    config = config.pipe(ops.first())

    kafka_request = config.pipe(
        ops.map(lambda c: create_model_topics(c)),
    )

    return ModelPublisherSink(
        file=file.Sink(request=rx.merge(config_read_request)),
        http=http.Sink(request=http_request),
        kafka=kafka.Sink(request=kafka_request),
    )


def main():
    loop = asyncio.get_event_loop()
    aio_scheduler = AsyncIOScheduler(loop=loop)
    run(Component(call=functools.partial(model_publisher, aio_scheduler), input=ModelPublisherSource),
        ModelPublisherDrivers(
            kafka=kafka.make_driver(),
            http=http.make_driver(),
            file=file.make_driver(),
            argv=argv.make_driver(),
        )
    )
